import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np

from .baseclass import baseclass as _baseclass
import warnings as _warnings
from .conversions import *


class Scatter(_baseclass):
    """
    Represents a foil element that scatters the beam.

    Parameters
    ----------

    length : float
        Length of the element.
    thickness : float
        The thickness of the foil in SI units.
    radlength : float
        The radiation lenght of the foil.
    order : int
        Order to calculate the transfer matrix.
    name : str
        The name used to identify the element.
    verbose : boolean
        Whether or not to print information to the terminal
    """
    _type = 'scatter'
    _order = 1

    def __init__(self, thickness=0, radlength=1, name=None, verbose=False):
        self.name   = name
        self.verbose = verbose
        self._x      = thickness
        self._X0     = radlength

    def theta_rms(self, energy_GeV):
        """
        The RMS scattering angle in radians.
        """
        x_div_X0 = self._x/self._X0
        theta = 13.6e-3/energy_GeV * _np.sqrt(x_div_X0) * (1+0.038*_np.log(x_div_X0))
        return theta
    # theta_rms = property(_get_theta_rms)

    @property
    def R(self):
        """
        This is necessary for the simulation, but a scattering element does not have an R matrix, and this is the identity.
        """
        _warnings.warn('This scatter element does not have an R matrix: returning identity instead', UserWarning, stacklevel=3)
        return _np.identity(6)

    def change_E(self, old_gamma, new_gamma):
        """
        Changes the energy of the element.
        """
        pass

    @property
    def ele_name(self):
        """
        Returns the elegant-styled name.
        """
        name = 'SCATTER_{}_{ind:03.0f}'.format(self.name, self.ind)
        return name

    @property
    def ele_string(self, ind, gamma):
        """
        Returns the full elegant entry.
        """
        string = (
            '{name}\t:SCATTER , &\n'
            '\t\tXP = {xp}    , &\n'
            '\t\tYP = {yp}'
            ).format(
                name = self.elename,
                xp   = self.theta_rms(gamma2GeV(gamma)),
                yp   = self.theta_rms(gamma2GeV(gamma))
                )
        string = string + self._kwargs_str

        return string



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import numpy as _np
from .classes import Bunch

__all__ = [
    'gaussbunch_sigmas',
    'gaussbunch_twiss'
    ]


def gaussbunch_sigmas(sx, sxp, sy, syp, sz, sd, n_parts):
    """
    Generates a gaussian bunch with RMS moments.

    Parameters
    ----------

    sx : float
        RMS moment in :math:`x`.
    sxp : float
        RMS moment in :math:`x'`.
    sy : float
        RMS moment in :math:`y`.
    syp : float
        RMS moment in :math:`y'`.
    sz : float
        RMS moment in :math:`z'`.
    sd : float
        RMS moment in :math:`\\delta = \\frac{p - p_0}{p_0}`.

    Returns
    -------
    
    bunch : :class:`slactrac.Bunch`.
    """
    x  = _np.random.normal(loc=0, scale=sx, size=n_parts)
    xp = _np.random.normal(loc=0, scale=sxp, size=n_parts)

    y  = _np.random.normal(loc=0, scale=sy, size=n_parts)
    yp = _np.random.normal(loc=0, scale=syp, size=n_parts)

    z  = _np.random.normal(loc=0, scale=sz, size=n_parts)
    d  = _np.random.normal(loc=0, scale=sd, size=n_parts)

    return Bunch(x, xp, y, yp, z, d)


def gaussbunch_twiss(Beam_x, Beam_y, sz, sd, n_parts, szd=0):
    """
    Generates a gaussian bunch with:

    Parameters
    ----------

    Beam_x : float
        :class:`slactrac.BeamParams` representation of beam properties in x.
    Beam_y : float
        :class:`slactrac.BeamParams` representation of beam properties in y.
    sz : float
        RMS moment in :math:`z'`.
    sd : float
        RMS moment in :math:`\\delta = \\frac{p - p_0}{p_0}`.

    Returns
    -------
    
    bunch : :class:`slactrac.Bunch`.
    """
    mean = _np.zeros(6)
    cov_xx = Beam_x.beta*Beam_x.emit
    cov_xxp = Beam_x.avg_xxp
    cov_xpxp = Beam_x.gamma*Beam_x.emit

    cov_yy = Beam_y.beta*Beam_y.emit
    cov_yyp = Beam_y.avg_xxp
    cov_ypyp = Beam_y.gamma*Beam_y.emit

    cov = _np.array(
        [
            [cov_xx   , cov_xxp   , 0        , 0         , 0                 , 0              ]    ,
            [cov_xxp  , cov_xpxp  , 0        , 0         , 0                 , 0              ]    ,
            [0        , 0         , cov_yy   , cov_yyp   , 0                 , 0              ]    ,
            [0        , 0         , cov_yyp  , cov_ypyp  , 0                 , 0              ]    ,
            [0        , 0         , 0        , 0         , _np.power(sz, 2)  , szd            ]    ,
            [0        , 0         , 0        , 0         , szd               , _np.power(sd, 2) ]
        ]
        )

    parts = _np.random.multivariate_normal(mean, cov, size=n_parts)
    x = parts[:, 0]
    xp = parts[:, 1]

    y = parts[:, 2]
    yp = parts[:, 3]

    z = parts[:, 4]
    d = parts[:, 5]

    return Bunch(x, xp, y, yp, z, d)



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import slactrac as _sltr
    import numpy as _np
    pass


def _store_emit(self, emit=None, emit_n=None):
    if (emit is None) and (emit_n is None):
        raise TypeError('Either emit or emit_n must be specified.')
    elif (emit is not None) and (emit_n is not None):
        raise TypeError('Emit and emit_n cannot be specified simultaneously.')
    elif emit is not None:
        self._emit = emit
    elif emit_n is not None:
        self._emit = emit_n/_sltr.GeV2gamma(self.E)
    else:
        raise TypeError('Something has gone wrong.')


class BeamBase(object):
    def __init__(self, nb0, E=20.35, dE=0.01, emit=None, emit_n=None):
        self._store_defaults(nb0, E, dE)
        _store_emit(emit=emit, emit_n=emit_n)

    def _store_defaults(self, nb0, E, dE):
        self._nb0 = nb0
        self._E   = E
        self._dE  = dE

    @property
    def nb0(self):
        """
        On-axis beam density.
        """
        return self._nb0

    @property
    def E(self):
        """
        Beam energy in GeV.
        """
        return self._E

    @property
    def dE(self):
        """
        Beam energy spread.
        """
        return self._dE

    @property
    def emit(self):
        """
        Beam emittance :math:`\\epsilon`.
        """
        return self._emit

    @property
    def emit_n(self):
        """
        Normalized beam emittance :math:`\\gamma \\epsilon`.
        """
        return self._emit*_sltr.GeV2gamma(self.E)


class RoundBeam(BeamBase):
    """
    Creates a beam with circularly-symmetric properties.

    Parameters
    ----------

    nb0 : float
        On-axis beam density.
    s_r : float
        The beam RMS width :math:`\\sigma_r`.
    E : float
        Beam energy in GeV.
    dE : float
        Beam energy spread.
    """
    def __init__(self, nb0, s_r, E=20.35, dE=0.01):
        super().__init__(nb0=nb0, E=E, dE=dE)
        self._s_r = s_r

    @property
    def s_r(self):
        """
        The beam RMS width :math:`\\sigma_r`.
        """
        return self._s_r


class EllipseBeam(BeamBase):
    """
    Creates an ellipsoidal beam.
    
    Either the set of parameters `sx`, `sxp`, and `sxxp` must be used; or `beta`, `alpha`, and `emit` or `emit_n` must be used.

    Parameters
    ----------

    nb0 : float
        On-axis beam density.
    E : float
        Beam energy in GeV.
    dE : float
        Beam energy spread.
    sx : float, optional
        Beam moment where :math:`\\text{sx}^2 = \\langle x^2 \\rangle`.
    sxp : float, optional
        Beam moment where :math:`\\text{sxp}^2 = \\langle x'^2 \\rangle`.
    sxxp : float, optional
        Beam moment where :math:`\\text{sxxp} = \\langle x x' \\rangle`.
    beta : float, optional
        Courant-Snyder parameter :math:`\\beta`.
    alpha : float, optional
        Courant-Snyder parameter :math:`\\alpha`.
    emit : float, optional
        Beam emittance :math:`\\epsilon`.
    emit_n : float, optional
        Normalized beam emittance :math:`\\gamma \\epsilon`.
    """
    def __init__(self, nb0, E=20.35, dE=0.01, sx=None, sxp=None, sxxp=None, beta=None, alpha=None, emit=None, emit_n=None):
        self._store_defaults(nb0, E, dE)
        try:
            self.set_Courant_Snyder(beta=beta, alpha=alpha, emit=emit, emit_n=emit_n)
        except:
            self.set_moments(sx=sx, sxp=sxp, sxxp=sxxp)

    def set_moments(self, sx, sxp, sxxp):
        """
        Sets the beam moments directly.

        Parameters
        ----------
        sx : float
            Beam moment where :math:`\\text{sx}^2 = \\langle x^2 \\rangle`.
        sxp : float
            Beam moment where :math:`\\text{sxp}^2 = \\langle x'^2 \\rangle`.
        sxxp : float
            Beam moment where :math:`\\text{sxxp} = \\langle x x' \\rangle`.
        """
        self._sx   = sx
        self._sxp  = sxp
        self._sxxp = sxxp
        emit = _np.sqrt(sx**2 * sxp**2 - sxxp**2)
        self._store_emit(emit=emit)

    @property
    def sx(self):
        """
        Beam moment where :math:`\\text{sx}^2 = \\langle x^2 \\rangle`.
        """
        return self._sx

    @property
    def sxp(self):
        """
        Beam moment where :math:`\\text{sxp}^2 = \\langle x'^2 \\rangle`.
        """
        return self._sxp

    @property
    def sxxp(self):
        """
        Beam moment where :math:`\\text{sxxp} = \\langle x x' \\rangle`.
        """
        return self._sxxp

    def set_Courant_Snyder(self, beta, alpha, emit=None, emit_n=None):
        """
        Sets the beam moments indirectly using Courant-Snyder parameters.

        Parameters
        ----------
        beta : float
            Courant-Snyder parameter :math:`\\beta`.
        alpha : float
            Courant-Snyder parameter :math:`\\alpha`.
        emit : float
            Beam emittance :math:`\\epsilon`.
        emit_n : float
            Normalized beam emittance :math:`\\gamma \\epsilon`.
        """

        self._store_emit(emit=emit, emit_n=emit_n)
        
        self._sx   = _np.sqrt(beta*self.emit)
        self._sxp  = _np.sqrt((1+alpha**2)/beta*self.emit)
        self._sxxp = -alpha*self.emit

    @property
    def beta(self):
        """
        Courant-Snyder parameter :math:`\\beta`.
        """
        beta = _np.sqrt(self.sx)/self.emit
        return beta

    @property
    def alpha(self):
        """
        Courant-Snyder parameter :math:`\\alpha`.
        """
        alpha = -self.sxxp/self.emit
        return alpha



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np


class BeamParams(object):
    """
    Calculates Courant-Snyder parameters and beam properties in a self-consistent manner.

    Parameters
    ----------

    beta : float
        Beta function value in meters.
    alpha : float
        Alpha function value.
    emit : float
        Emittance in SI units.
    emit_n : float
        Normalized emittance in SI units.
    gamma_en : float
        Relativistic gamma :math:`\\gamma`
    spotsize : float
        RMS of :math:`x` coordinate.
    divergence : float
        RMS of :math:`x'` coordinate.
    r : float
        The average correlation of :math:`x` and :math:`x'`.
    avg_xxp : float
        The average correlation of :math:`x` and :math:`x'`.
    """
    def __repr__(self):
        return '<{} at {}; beta={}, alpha={}>'.format(self.__class__.__module__, hex(id(self)), self.beta, self.alpha)

    def __init__(self,
            beta=None, alpha=None, emit=None, emit_n=None, gamma_en=None,
            spotsize=None, divergence=None, r=None, avg_xxp=None):
        # If not enough info for Twiss
        emit_norm_insufficient = (emit_n is None) or (gamma_en is None)
        emit_geom_insufficient = emit is None
        emit_insufficient      = (emit_geom_insufficient and emit_norm_insufficient)
        shape_insufficient     = (beta is None) or (alpha is None)
        twiss_insufficient     = (shape_insufficient or emit_insufficient)
        moments_insufficient   = (spotsize is None) or (divergence is None) or ( (r is None) and (avg_xxp is None) )

        # Validation that there's enough info to construct everything
        if twiss_insufficient and moments_insufficient:
            raise ValueError('Not enough information to construct Twiss values')

        if twiss_insufficient:
            if avg_xxp is not None:
                emit = _np.sqrt(spotsize**2 * divergence**2 - avg_xxp**2)
                corr_sign = _np.sign(avg_xxp)
            elif r is not None:
                emit = spotsize * divergence * _np.sqrt(1-r**2)
                corr_sign = _np.sign(r)
            else:
                raise ValueError('Missing avg_xxp/r term')
            beta = spotsize**2/emit
            gamma = divergence**2/emit
            alpha = -corr_sign * _np.sqrt(gamma*beta-1)
        else:
            def _get_geom_emit(emit=None, emit_n=None, gamma=None):
                if emit is None and emit_n is None:
                    raise ValueError('Did not specify an emittance!')
                if emit is not None and emit_n is not None:
                    raise ValueError('Specified too many emittance values!')
                if emit_n is not None:
                    # print emit_n
                    emit = _np.float64(emit_n)/_np.float64(gamma)
                else:
                    emit = _np.float64(emit)
                return emit
            emit = _get_geom_emit(emit=emit, emit_n=emit_n, gamma=gamma_en)

        self.beta  = _np.float64(beta)
        self.alpha = _np.float64(alpha)
        self.emit  = _np.float64(emit)

    def set_emit_n(self, emit_n, gamma):
        """
        Set the normalized emittance of the beam.
        """
        self.emit = _np.float64(emit_n/gamma)

    # Definte beta property
    # Validate beta > 0
    @property
    def beta(self):
        """
        The beta function of the beam
        """
        return self._beta

    @beta.setter
    def beta(self, value):
        if not (value > 0):
            raise ValueError('Beta must be greater than zero: requested beta={}.'.format(value))
        self._beta = value

    @property
    def betastar(self):
        """
        The minimum beta function of the beam in vacuum.
        """
        return 1.0/self.gamma

    @property
    def sstar(self):
        """
        The distance from the minimum beta function of the beam in vacuum.
        """
        return self.alpha/self.gamma

    # Define gamma property
    # Derived from beta and alpha
    @property
    def gamma(self):
        """
        The gamma function of the beam
        """
        return (1+_np.power(self.alpha, 2))/self.beta

    # Define T matrix
    # Derived from beta, alpha, and gamma
    @property
    def T(self):
        """
        The beam T matrix.
        """
        return _np.array([[self.beta, -self.alpha], [-self.alpha, self.gamma]])

    # Define transport method
    # Returns new twiss given R matrix.
    def transport(self, R):
        """
        Given a transport matrix *R*, return the new beam parameters.

        Returns :class:`slactrac.BeamParams`.
        """
        T2 = _np.dot(_np.dot(R, self.T), _np.transpose(R))
        betaf = T2[0, 0]
        alphaf = -T2[0, 1]
        return BeamParams(beta=betaf, alpha=alphaf, emit=self.emit)

    # Define spotsize method
    @property
    def spotsize(self):
        """
        The RMS size of the beam.
        """
        return _np.sqrt(self.beta*self.emit)

    @property
    def divergence(self):
        """
        The divergence of the beam.
        """
        return _np.sqrt(self.gamma*self.emit)

    @divergence.setter
    def divergence(self, value):
        self.emit = value**2/self.gamma

    @property
    def avg_xxp(self):
        """
        The average of coordinates :math:`x` and :math:`x'`.
        """
        return -self.alpha*self.emit

    @property
    def minspotsize(self):
        """
        The RMS size of the beam at its smallest in vacuum.
        """
        return _np.sqrt(self.betastar*self.emit)



import jinja2 as _jj
import os as _os
import pkg_resources as _pkg_resources
import shlex as _shlex
import shutil as _shutil
import tempfile as _tempfile
import subprocess as _subprocess

import logging as _logging
_logger      = _logging.getLogger(__name__)
_loggerlevel = _logging.DEBUG
#  _loggerlevel = 9

__all__ = ['elegant_sim']


def elegant_sim(beamline, dir=None, filename=None, **kwargs):
    """
    Creates and runs a simulation in Elegant.


    Parameters
    ----------

    dir : str
        If none, simulated in a :func:`tempfile.mkdtemp` temporary directory.
    filename : str
        If none, simulated with a :func:`tempfile.mkstemp` temporary file.

    Returns
    -------

    (path, root, ext) : (str, str, str)
        A tuple of the path of the simulation, the filename of the simulation, and the extension of the simulation
    """
    # ======================================
    # If path isn't specified, create
    # temporary one
    # ======================================
    if dir is None:
        path = _tempfile.mkdtemp()
        _logger.log(level=_loggerlevel, msg='Path is: {}'.format(path))
    else:
        path = dir

    # ======================================
    # Open up a temporary file in the path
    # ======================================
    if filename is None:
        fd, filename = _tempfile.mkstemp(dir=path, prefix='out_', suffix='.ele')
        f = _os.fdopen(fd, 'w+')
    else:
        f = open(_os.path.join(path, filename), 'w+')
    _logger.log(level=_loggerlevel, msg='Filename is: {}'.format(filename))

    basename = _os.path.basename(filename)
    root, ext = _os.path.splitext(basename)
    lte_name = _os.path.join(path, '{}.lte'.format(root))
    beamline.elegant_lte(filename=lte_name)
    
    # ======================================
    # Get template path from package and
    # load template
    # ======================================
    templates_path = _pkg_resources.resource_filename(__name__, 'templates')
    loader = _jj.FileSystemLoader(templates_path)
    env = _jj.Environment(loader=loader)
    template = env.get_template('template.ele')
    
    # ======================================
    # Generate ele file
    # ======================================
    template.stream(filename=lte_name, matched=0, **kwargs).dump(f)
    f.close()

    # ======================================
    # Run Elegant
    # ======================================
    command = 'elegant {}'.format(filename)
    _logger.log(level=_loggerlevel, msg='Command is: {}'.format(command))

    log_name = _os.path.join(path, '{}.log'.format(root))
    f_log = open(log_name, 'w+')
    _logger.log(level=_loggerlevel, msg='Log file is: {}'.format(log_name))

    #  fnull = open(_os.devnull, 'w')
    cwdu = _os.getcwdu()
    _os.chdir(path)
    _subprocess.call(_shlex.split(command), stdout=f_log, stderr=f_log)
    _os.chdir(cwdu)

    # ======================================
    # Clean temporary directory if necessary
    # ======================================
    if dir is None:
        _shutil.rmtree(path)

    return path, root, ext



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import slactrac as _sltr
    import numpy as _np
    pass

from .plasma import Plasma as _Plasma
from .particles import GaussPartBeam as _GaussPartBeam
import scipy.constants as spc


class SimBeam(object):
    """
    Simulates beam particles in a plasma, given certain initial conditions.

    Parameters
    ----------

    plasma : 
    E0 : float
        Mean beam energy.
    n_p_cgs : float
        Plasma density in CGS units.
    nparts : int
        Number of particles to use in simulation
    beta_mismatch : float
        Factor to mismatch distribution in beta by.
    s_pts : int
        Number of points to simulate in s.
    """
    def __init__(self, plasma, beam, E0, n_p_cgs, nparts, beta_mismatch, s_pts):
        self._E0            = E0
        self._n_p_cgs       = n_p_cgs
        self._nparts        = nparts
        self._beta_mismatch = beta_mismatch
        self._s_pts         = s_pts
        self._plasma        = plasma
        self._beam          = beam

        # ==================================
        # Set up coordinate arrays
        # ==================================
        s = _np.linspace(0, 20*2*_np.pi*self.plasma.beta_matched(E0), s_pts)
        self._s = s
        y = _np.empty([nparts, s.shape[0], 2])
        self._phi = _np.empty([nparts, s.shape[0]])
        
        # ==================================
        # Defaults of motion
        # ==================================
        k0 = self.plasma.k_ion(E0)
        rtk0 = _np.sqrt(k0)

        k_xi = _np.sqrt(spc.elementary_charge**2 * self._n_p_cgs / (4*_np.pi*spc.epsilon_0*spc.electron_mass*spc.speed_of_light**2) ) / _np.cos(self.beam.xi) 

        self._k_xi = k_xi
        
        # ==================================
        # Get particle coordinates
        # ==================================
        for i, delta in enumerate(self.beam.delta):
            E = self.beam.E * (1+delta)
            plas = _Plasma(n_p_cgs=self._n_p_cgs/(_np.cos(k_xi[i] * self.beam.xi[i])**2))
            k = plas.k_ion(E)
            # k = self.plasma.k_ion(E)
            rtk = _np.sqrt(k)
            x0 = self.beam.x[i]
            xp0 = self.beam.xp[i]
            # y0 = [beam.xp[i], beam.x[i]]
            # y[i, :, :] = sp.integrate.odeint(self.ion_force, y0, s, args=(E,))
            y[i, :, 0] = -x0*rtk*_np.sin(rtk*s) + xp0*_np.cos(rtk*s)
            y[i, :, 1] = x0*_np.cos(rtk*s) + xp0*_np.sin(rtk*s)/rtk
            self._phi[i, :] = _np.mod((rtk-rtk0)*s+_np.pi, 2*_np.pi) - _np.pi
        
        # ==================================
        # Extract particle coordinates
        # ==================================
        self._x = y[:, :, 1]
        self._xp = y[:, :, 0]

    @property
    def k_xi(self):
        """
        Ion focusing wavenumber :math:`k_\\xi`.
        """
        return self._k_xi


    @property
    def phi(self):
        """
        Particle phases :math:`\\phi`.
        """
        return self._phi

    @property
    def beam(self):
        """
        Initial beam object.
        """
        return self._beam

    @property
    def plasma(self):
        """
        Initial plasma object.
        """
        return self._plasma

    @property
    def s(self):
        """
        Coordinates of beam (:math:`s`).
        """
        return self._s
    
    @property
    def x(self):
        """
        Coordinates of beam (:math:`x`).
        """
        return self._x

    @property
    def xp(self):
        """
        Coordinates of beam (:math:`x'`).
        """
        return self._xp

    # ==================================
    # Integrate particles
    # ==================================
    def ion_force(self, y, t, E):
        xp = y[0]
        x  = y[1]
        return [-self.plasma.k_ion(E)*x, xp]

    @property
    def spotsq(self):
        """
        The beam variance :math:`\\langle x^2 \\rangle`.
        """
        return _np.mean(self.x**2, axis=0)

    @property
    def divsq(self):
        """
        The beam divergence :math:`\\langle x'^2 \\rangle`.
        """
        return _np.mean(self.xp**2, axis=0)

    @property
    def xxp(self):
        """
        The beam correlation :math:`\\langle x x' \\rangle`.
        """
        return _np.mean(self.x*self.xp, axis=0)
    
    @property
    def emit_measured(self):
        """
        The beam emittance :math:`\\langle x x' \\rangle`.
        """
        return _np.sqrt(self.spotsq*self.divsq-self.xxp**2)



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np

from .driftmat import driftmat as _driftmat
from .baseclass import baseclass as _baseclass

__all__ = ['Bend']


class Bend(_baseclass):
    """
    Represents a bend element.
    
    Parameters
    ----------
    
    length : float
        The length of the magnet.
    angle : float
        The bend angle of the magnet.
    rotate : float
        The rotation about :math:`z` of the magnet.
    order : float
        Order to calculate matrix elements.
    name : str
        Unique name to identify the element.
    """
    def __init__(self, length=0, angle=0, order=1, rotate=0, name=None, **kwargs):
        self.name   = name
        self._type   = 'bend'
        self._length = _np.float_(length)
        self._order  = int(order)
        self._angle  = _np.float_(angle)
        self.rotate  = rotate
        self._kwargs = kwargs

    @property
    def rotate(self):
        """
        The rotation about :math:`z` of the magnet.
        """
        return self._rotate

    @rotate.setter
    def rotate(self, val):
        self._rotate = val

    @property
    def R(self):
        """
        The first-order transfer matrix.
        """
        temp = bendmat(
            length = self._length,
            angle  = self._angle,
            order  = self._order
            )
        if self.rotate == 90:
            R          = _np.zeros([6, 6])
            R[0:2, 0:2] = temp[2:4, 2:4]
            R[2:4, 0:2] = temp[0:2, 2:4]
            R[0:2, 5]   = temp[2:4, 5]
            R[0:2, 2:4] = temp[2:4, 0:2]
            R[2:4, 2:4] = temp[0:2, 0:2]
            R[2:4, 5]   = temp[0:2, 5]
            R[4:6, 0:2] = temp[4:6, 2:4]
            R[4:6, 2:4] = temp[4:6, 0:2]
            R[4:6, 4:6] = temp[4:6, 4:6]
            # print('hi')
        elif self.rotate == 0:
            R = temp
        else:
            print('SERIOUS PROBLEM')
            R = temp
            # print('hi')
        return R

    @property
    def length(self):
        """
        The length of the magnet.
        """
        return self._length

    @property
    def angle(self):
        """
        The bend angle of the magnet in degrees.
        """
        return self._angle

    @property
    def tilt(self):
        """
        The bend angle of the magnet in radians.
        """
        return self.rotate*_np.pi/180

    def change_E(self, old_gamma, new_gamma):
        """
        Scale the setpoint energy of the magnet from the old energy *old_gamma* to new energy *new_gamma*.

        The best way to think about this is that the angle will change with different beam energy, so changing the beam energy changes the magnet's properties, because the magnet's B-field stays the same.
        """
        old_gamma = _np.float_(old_gamma)
        new_gamma = _np.float_(new_gamma)
        self._angle *= old_gamma / new_gamma

    @property
    def ele_name(self):
        """
        Returns the elegant-styled name.
        """
        name = 'BEND_{}_{:03.0f}'.format(self.name, self.ind)
        return name

    @property
    def ele_string(self):
        """
        Returns the full elegant entry.
        """
        string = (
            '{}\t:CSRCSBEN               , &\n'
            '\t\tL            = {:0.10e} , &\n'
            '\t\tANGLE        = {}       , &\n'
            '\t\tTILT         = {}       , &\n'
            '\t\tN_KICKS      = 20       , &\n'
            '\t\tSG_HALFWIDTH = 1        , &\n'
            '\t\tSG_ORDER     = 1        , &\n'
            '\t\tSTEADY_STATE = 0        , &\n'
            '\t\tBINS         = 500').format(self.ele_name, self.length, self.angle, self.tilt)
        string = string + self._kwargs_str

        return string


def bendmat(
        length=0,
        angle=0,
        order=1,
        fse=0,
        K1=0,
        K2=0
        ):

    if (( length == 0 ) or ( angle == 0 )):
        return _driftmat(length, order)

    # Make sure ints don't flow through
    length = _np.float_(length)
    angle  = _np.float_(angle)
    K1     = _np.float_(K1)
    K2     = _np.float_(K2)

    h     = angle/length            # /* coordinate system curvature */
    ha    = h*(1+fse)               # /* actual curvature due to bending field */
    n     = -K1/_np.square(h)              # /* field index */
    beta  = K2/2/pow(h, 3)            # /* sextupole index */
    gamma = 0

    t0      = length
    nh      = n*h
    betah2  = beta*_np.square(h)  # noqa
    gammah3 = gamma*pow(h, 3)  # noqa
    
    # M = sbend_matrix(length, h, ha, n*h, beta*sqr(h), gamma*pow3(h), order)
    h2 = h*h
    ha2 = ha*ha  # noqa

    kx2 = -(h2 - 2*h*ha + nh*ha)
    ky2 = nh*ha
    ky2_is_zero = (ky2 == 0)  # noqa
    kx2_is_zero = (kx2 == 0)  # noqa

    small3 = pow(1e-16, 1./3.)

    if (_np.sqrt(abs(kx2))*t0 < small3):
        kx2 = _np.square(small3/t0)
        if (kx2):
            kx2 = _np.sign(kx2)*kx2

    if (kx2 > 0.0):
            kx = _np.sqrt(kx2)
            cx = _np.cos(kx*t0)
            sx = _np.sin(kx*t0)/kx
    elif (kx2 < 0.0):
            kx = _np.sqrt(-kx2)
            cx = _np.cosh(kx*t0)
            sx = _np.sinh(kx*t0)/kx

    if (abs(ky2) < small3):
        ky2 = _np.square(small3/t0)
        if (ky2):
            ky2 = _np.sign(ky2)*ky2

    if (ky2 > 0.0):
            ky = _np.sqrt(ky2)
            cy = _np.cos(ky*t0)
            sy = _np.sin(ky*t0)/ky
    elif (ky2 < 0.0):
            ky = _np.sqrt(-ky2)
            cy = _np.cosh(ky*t0)
            sy = _np.sinh(ky*t0)/ky

    R = _np.identity(6)
    if (h != 0):
        R[0][0] = R[1][1] = cx
        R[0][1] = sx
        R[0][2] = R[0][3] = R[0][4] = 0
        R[0][5] = -ha*(cx-1)/kx2
        R[1][0] = -sx*kx2
        R[1][2] = R[1][3] = R[1][4] = 0
        R[1][5] = sx*ha
        R[2][2] = R[3][3] = cy
        R[2][3] = sy
        R[2][0] = R[2][1] = R[2][4] = R[2][5] = 0
        R[3][2] = - sy*ky2
        R[3][0] = R[3][1] = R[3][4] = R[3][5] = 0
        R[4][0] = h*sx
        R[4][1] = -(h*(cx - 1))/kx2
        R[4][2] = R[4][3] = 0
        R[4][5] = (h*ha*(t0 - sx))/kx2
    else:
        # /* steering corrector */
        R[0][0] = R[1][1] = R[2][2] = R[3][3] = 1
        R[0][1] = R[2][3] = t0
        R[0][5] = C[0] + t0*_np.sin(ha*t0)  # noqa
        R[1][5] = t0*ha*_np.cos(ha*t0)

    R[5][0] = R[5][1] = R[5][2] = R[5][3] = R[5][4] = 0
    R[4][4] = R[5][5] = 1
 
    return R



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import scipy.constants as _spc
    import numpy as _np
    import scipy as _sp

    # ============================
    # Constants
    # ============================
    e      = _spc.e               # Elementary charge
    c      = _spc.speed_of_light  # Speed of light
    e0     = _spc.epsilon_0       # Vacuum permittivity
    amu    = _spc.atomic_mass     # AMU in kg

from .ions import Ions as _Ions

import logging as _logging
_logger = _logging.getLogger(__name__)


class Ions2D(_Ions):
    """
    A class to facilitate calculating ion motion in PWFA ion columns due to cylindrical, infinitely-long gaussian beams.

    .. versionadded:: 1.6

    Parameters
    ----------

    species : :class:`periodictable.core.Element`
        The plasma gas species, e.g. :class:`periodictable.H`.
    N_e : float
        The number of electrons.
    sig_r : float
        The standard deviation in :math:`r`.
    sig_xi : float
        The standard deviation in :math:`\\xi`.
    r0_big : float
        The maximum particle coordinate in :math:`r` to include.
    n_samp : int
        The number of samples.
    order : int
        The order to calculate to
    rtol : float
        A tolerance to calculate to. See :func:`scipy.integrate.odeint`.
    atol : float
        A tolerance to calculate to. See :func:`scipy.integrate.odeint`.
    """
    def __init__(self, species, N_e, sig_r, sig_xi, r0_big=None, n_samp=1000, order=5, rtol=None, atol=None):
        # ============================
        # Experiment Parameters
        # ============================
        super().__init__(dims=2, species=species, N_e=N_e, sig_xi=sig_xi, rtol=rtol, atol=atol)
        self._sig_r  = sig_r   # Transverse R.M.S. width

    def _basic_shape(self, r0_big=None, order=4, n_samp=None):
        if r0_big is None:
            r0_big = self.sig_r * 10000

        if n_samp is None:
            n_samp = 1e4+1
        # ============================
        # Get basic shape
        # (Fourier series)
        # ============================

        x_end = self.lambda_large(r0_big)
        dx = x_end / (n_samp-1)
        x = _np.linspace(0, x_end, n_samp)
        
        y_num = self.r(x, r0_big)
        f_x = y_num / r0_big
        
        a = _np.zeros(order+1)
        for i in range(order+1):
            if i % 2 == 1:
                a[i] = _np.sum(f_x * self.r_large(i*x, r0_big))/r0_big * dx * 2/self.lambda_large(r0_big)

        out = _np.zeros(x.shape)
        for i, val in enumerate(a):
            out = out + val*self.r_large(i*x, r0_big)

        print('=======================')
        print('Initial conditions')
        print('=======================')
        print('k     = {}'.format(self.k))
        print('sig_r = {}'.format(self.sig_r))
        print('r0    = {}'.format(r0_big))
        print('a     = {}'.format(a))

        return x, y_num, out

    @property
    def nb0(self):
        """
        On-axis beam density :math:`n_{b,0}`.
        """
        return self.N_e / ( (2*_np.pi)**(3/2) * self.sig_r**2 * self.sig_xi)

    def lambda_large(self, r0):
        """
        The wavelength for large (:math:`r_0 < \\sigma_r`) oscillations.
        """
        return 2*_np.sqrt(2*_np.pi/self.k)*r0

    def r_small(self, x, r0):
        """
        Approximate trajectory function for small (:math:`r_0 < \\sigma_r`) oscillations.
        """
        return r0*_np.cos(_np.sqrt(self.k_small) * x)

    def r_large(self, x, r0):
        """
        Approximate trajectory function for large (:math:`r_0 > \\sigma_r`) oscillations.
        """
        return r0*_np.cos(x*self.omega_big(r0))

    @property
    def k(self):
        """
        Driving force term: :math:`r'' = -k \\left( \\frac{1-e^{-r^2/2{\\sigma_r}^2}}{r} \\right)`
        """
        try:
            return self._k
        except AttributeError:
            self._k = e**2 * self.N_e / ( (2*_np.pi)**(5/2) * e0 * self.m * c**2 * self.sig_xi)
            return self._k

    @property
    def k_small(self):
        """
        Small-angle driving force term: :math:`r'' = -k_{small} r`.

        Note: :math:`k_{small} = \\frac{k}{2{\\sigma_r^2}}`
        """
        return self.k / (2*self.sig_r**2)

    @property
    def sig_r(self):
        """
        Transverse R.M.S. width
        """
        return self._sig_r

    # ============================
    # Force equation
    # ============================
    def _rp2(self, r):
        if r == 0:
            return 0
        return -self.k*(1-_np.exp(-r**2/(2*self.sig_r**2)))/r
    
    # ============================
    # Derivative of force eq.
    # ============================
    def _delrp2(self, r):
        return (1/r**2-_np.exp(-r**2/(2*self.sig_r**2))*(1/r**2+1/self.sig_r**2)) * self.k

    # ============================
    # First-order ODE system
    # ============================
    def _func(self, y, t):
        return [self._rp2(y[1]), y[0]]

    # ============================
    # Gradient for 1st-order ODE
    # ============================
    def _gradient(self, y, t):
        return [[0, self._delrp2(y[1])], [1, 0]]

    # ============================
    # Scale factor function
    # ============================
    def h(self, q):
        return _np.abs(q)

    @property
    def q_label(self):
        return 'r'



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
from .beamparams import BeamParams as _BeamParams
from .baseclass import baseclass as _baseclass
from .conversions import gamma2GeV as _gamma2GeV
import copy as _copy
import logging as _logging
import warnings as _warnings

_logger = _logging.getLogger(__name__)
#  _loggerlevel = _logging.DEBUG
_loggerlevel = 9

#  class Elements(object):
#      def __init__(self, element_list, verbose=None):
#          self._elements


class Beamline(_baseclass):
    """
    Represents a beamline composed of various elements.

    Parameters
    ----------

    element_list : float
        A list of all the elements in order in the beamline.
    gamma : float
        The beamline energy.
    beam_x : float
        :class:`slactrac.Beamparams` class representing the x coordinates.
    beam_y : float
        :class:`slactrac.Beamparams` class representing the y coordinates.
    verbose : float
        If true, prints details of beam setup on the terminal.
    """
    _type    = 'beamline'
    _order   = int(1)

    def __init__(self, element_list, gamma, beam_x=None, beam_y=None, verbose=False):
        self.verbose = verbose
        self._gamma   = _np.float64(gamma)
        self.beam_x   = _copy.deepcopy(beam_x)
        self.beam_y   = _copy.deepcopy(beam_y)
        self.elements = _np.array([])

        for element in element_list:
            self.elements = _np.append(self.elements, _copy.deepcopy(element))
        for i, element in enumerate(self.elements):
            # element._gamma = self._gamma
            if (element._order > self._order):
                self._order = element._order
            if element.name is None:
                _warnings.warn('Missing name for {}-th element of beamline element list'.format(i), SyntaxWarning, stacklevel=2)
            element.ind = i

    def change_verbose(self, verbose=False):
        """
        Change whether the beamline is verbose.
        """
        for i, element in enumerate(self.elements):
            self.elements[i].verbose = verbose

    # Define property length
    @property
    def length(self):
        """
        The length of the beamline.
        """
        length = 0
        for element in self.elements:
            length += element._length
        return length

    # Define transfer matrix R property
    @property
    def R(self):
        """
        The transfer matrix R for the beamline.
        """
        R = _np.identity(6)
        for i, element in enumerate(self.elements):
            R = _np.dot(element.R, R)
        return R

    # Define gamma property for beamlines
    # If you change gamma, it changes for all elements
    @property
    def gamma(self):
        """
        The energy setpoint of the beamline.
        """
        return self._gamma

    @gamma.setter
    def gamma(self, gamma):
        gamma = _np.float64(gamma)
        for element in self.elements:
            # Changes energy of element
            element.change_E(self._gamma, gamma)
        self._gamma = gamma

    def _get_beam_end(self, beam, slice_var):
        beamtemp = _copy.deepcopy(beam)
        R = _np.identity(6)
        for i, element in enumerate(self.elements):
            if element._type == 'scatter':
                if self.verbose & element.verbose:
                    beamtemp   = beamtemp.transport(R[slice_var, slice_var])
                    old_emit_n = beamtemp.emit*self.gamma
                    theta_rms  = element.theta_rms(_gamma2GeV(self.gamma))
                    newdiv     = _np.sqrt(beamtemp.divergence**2 + theta_rms**2)
                    print('\t------------------------------------------------')
                    print('\tScatter element')
                    print('\tElement name: {}'.format(element.name))
                    print('\tDivergence:\n\t\tOld divergence:\t\t{:.3e}\n\t\tAdded Divergence:\t{:.3e}\n\t\tNew Divergence:\t\t{:.3e}\n\t\tFraction change:\t{:.3e}'.format(beamtemp.divergence, theta_rms, newdiv, newdiv/beamtemp.divergence-1))
                    beamtemp   = _BeamParams(spotsize=beamtemp.spotsize, divergence=newdiv, avg_xxp=beamtemp.avg_xxp)
                    new_emit_n = beamtemp.emit*self.gamma
                    print('\tEmittance:\n\t\tOld emittance:\t\t{}\n\t\tNew emittance:\t\t{}\n\t\tFraction change:\t{:.3e}'.format(old_emit_n, new_emit_n, new_emit_n/old_emit_n-1))
                    print('\t------------------------------------------------')
                    R = _np.identity(6)
            else:
                R = _np.dot(element.R, R)

        out = beamtemp.transport(R[slice_var, slice_var])
        return out

    @property
    def beam_x_end(self):
        """
        The beam parameters in x at the end of the beamline.
        """
        slice_var = slice(0, 2)
        out = self._get_beam_end(self.beam_x, slice_var)
        return out

    @property
    def beam_y_end(self):
        """
        The beam parameters in y at the end of the beamline.
        """
        slice_var = slice(2, 4)
        out = self._get_beam_end(self.beam_y, slice_var)
        return out

    @property
    def spotsize_x_end(self, emit=None, emit_n=None):
        """
        The beam spotsize in x at the end of the beamline.
        """
        return self.beam_x_end.spotsize

    @property
    def spotsize_y_end(self):
        """
        The beam spotsize in x at the end of the beamline.
        """
        if self.verbose:
            print('================================================')
            print('Verbose Beamline Printing in _get_spotsize_y_end...')
            print('================================================')
            print('Starting norm. emittance: {}'.format(self.beam_y.emit*self.gamma))
            print('Starting spot in y: {}'.format(self.beam_y.spotsize))
            print('Starting emittance: {}'.format(self.beam_y.emit))
            print('Starting beta: {}'.format(self.beam_y.beta))
        out = self.beam_y_end.spotsize
        if self.verbose:
            print('\t------------------------------------------------')
            print('\tFinal spot size: {:.3f} mm'.format(out*1e3))
            print('\t------------------------------------------------')
        return out

    def elegant_lte(self, filename='out.lte', scatter_enable=False):
        """
        Writes the beamline to an Elegant lte lattice file with *filename*.

        * *scatter_enable*: Determines if scatter elements are used.
        """
        f = open(filename, 'w')
        string = 'C\t:CHARGE,TOTAL=3.2E-09'
        f.write(string + '\n')
        names = _np.array('C')

        # Write elements
        for i, obj in enumerate(self.elements):
            if obj._type == 'scatter':
                if scatter_enable:
                    string = obj.ele_string(i, self.gamma)
                    names = _np.append(names, obj.ele_name)
                else:
                    continue
            else:
                string = obj.ele_string
                names = _np.append(names, obj.ele_name)

            _logger.log(level=_loggerlevel, msg='Name is: {}'.format(obj.ele_name))
            _logger.log(level=_loggerlevel, msg='Iteration is: {}'.format(i))
            f.write(string+'\n\n')

        # Write beamline
        string = '\nBEAMLINE\t:LINE=(&\n'
        for i, val in enumerate(names):
            if i == 0:
                string = string + '\t\t\t' + val
            else:
                string = string + ', &' + '\n\t\t\t' + val
        string = string + '  &\n\t\t\t)'

        f.write(string)

        f.close()



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
import warnings


class Twiss(object):
    def __repr__(self):
        return '<{} at {}; beta={}, alpha={}>'.format(self.__class__.__module__, hex(id(self)), self.beta, self.alpha)

    def __init__(self, beta, alpha):
        warnings.warn('Twiss is deprecated, use BeamParams instead', DeprecationWarning)
        self.beta  = _np.float64(beta)
        self.alpha = _np.float64(alpha)

    # Definte beta property
    # Validate beta > 0
    def _set_beta(self, value):
        if not (value > 0):
            raise ValueError('Beta must be greater than zero: requested beta={}.'.format(value))
        self._beta = value

    def _get_beta(self):
        return self._beta
    beta = property(_get_beta, _set_beta)
    
    def _get_betastar(self):
        return 1/self.gamma
    betastar = property(_get_betastar)

    def _get_sstar(self):
        return self.alpha/self.gamma
    sstar = property(_get_sstar)

    # Define gamma property
    # Derived from beta and alpha
    def _get_gamma(self):
        return (1+_np.power(self.alpha, 2))/self.beta
    gamma = property(_get_gamma)

    # Define T matrix
    # Derived from beta, alpha, and gamma
    def _get_T(self):
        return _np.array([[self.beta, -self.alpha], [-self.alpha, self.gamma]])
    T = property(_get_T)

    # Define transport method
    # Returns new twiss given R matrix.
    def transport(self, R):
        T2 = _np.dot(_np.dot(R, self.T), _np.transpose(R))
        betaf = T2[0, 0]
        alphaf = -T2[0, 1]
        return Twiss(betaf, alphaf)

    # Define spotsize method
    # Returns spot size given an emittance
    def spotsize(self, emit=None, emit_n=None, gamma=None):
        return _np.sqrt(self.beta*self._get_geom_emit(emit=emit, emit_n=emit_n, gamma=gamma))

    def _get_geom_emit(self, emit=None, emit_n=None, gamma=None):
        if emit is None and emit_n is None:
            raise ValueError('Did not specify an emittance!')
        if emit is not None and emit_n is not None:
            raise ValueError('Specified too many emittance values!')
        if emit_n is not None:
            # print emit_n
            emit = _np.float64(emit_n)/_np.float64(gamma)
        else:
            emit = _np.float64(emit)
        return emit

    def minspotsize(self, emit=None, emit_n=None, gamma=None):
        return _np.sqrt(self.betastar*self._get_geom_emit(emit=emit, emit_n=emit_n, gamma=gamma))



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import slactrac as _sltr
    import numpy as _np
    pass

from .beam import _store_emit


class GaussPartBeam(object):
    """
    Generates a Gaussian distribution of particles.

    Parameters
    ----------

    nparts : int
        Number of particles to use.
    q_tot : float
        Total charge of beam.
    E : float
        Average beam energy in GeV.
    sig_delta : float
        R.M.S. beam energy.
    beta : float
        Beam beta function.
    alpha : float
        Beam alpha function.
    emit : float
        Beam emittance.
    emit_n : float
        Normalized beam emittance.
    """
    def __init__(self, nparts, q_tot, E, sig_delta, sig_xi, beta, alpha, emit=None, emit_n=None):
        self._nparts    = _np.int(nparts)
        self._q_tot     = q_tot
        self._E         = E
        self._sig_delta = sig_delta
        self._sig_xi    = sig_xi
        self._beta      = beta
        self._alpha     = alpha
        _store_emit(self, emit=emit, emit_n=emit_n)

        cov      = _np.array([[self.beta, -self.alpha], [-self.alpha, (1+self.alpha**2)/self.beta]])*self.emit
        mean     = _np.array([0, 0])

        self._cov  = cov
        self._mean = mean

        coords   = _np.random.multivariate_normal(mean, cov, self.nparts)
        self._x  = coords[:, 0]
        self._xp = coords[:, 1]

        if sig_delta == 0:
            self._delta = _np.zeros(nparts)
        else:
            self._delta = _np.random.normal(scale=self.sig_delta, size=self.nparts)

        if sig_xi == 0:
            self._xi = _np.zeros(nparts)
        else:
            self._xi = _np.random.normal(scale=self.sig_xi, size=self.nparts)

    @property
    def sig_xi(self):
        """
        Std. dev. of :math:`\\xi`: :math:`\\sigma_\\xi`.
        """
        return self._sig_xi

    @property
    def xi(self):
        """
        Particle coordinates :math:`\\xi`.
        """
        return self._xi

    @property
    def mean(self):
        """
        Mean for x, x'.
        """
        return self._mean

    @property
    def cov(self):
        """
        Covariance for x, x'.
        """
        return self._cov

    @property
    def nparts(self):
        """
        Number of particles in the beam.
        """
        return self._nparts

    @property
    def x(self):
        """
        Particle coordinates :math:`x`.
        """
        return self._x

    @property
    def xp(self):
        """
        Particle coordinates :math:`x'`.
        """
        return self._xp

    @property
    def delta(self):
        """
        Particle coordinates :math:`\\delta`.
        """
        return self._delta

    @property
    def emit(self):
        """
        Beam emittance :math:`\\epsilon`.
        """
        return self._emit

    @property
    def E(self):
        """
        Beam energy in GeV.
        """
        return self._E

    @property
    def sig_delta(self):
        """
        Beam energy spread :math:`\\sigma_\\delta`.
        """
        return self._sig_delta

    @property
    def beta(self):
        """
        Beam beta :math:`\\beta`.
        """
        return self._beta

    @property
    def alpha(self):
        """
        Beam alpha :math:`\\alpha`.
        """
        return self._alpha



# Author Joel Frederico, SLAC National Accelerator Laboratory
"""
"""
__version__ = '1.1.5'
__all__ = [
    'BeamParams',
    'Beamline',
    'Bend',
    'Drift',
    'Focus',
    'GeV2gamma',
    'GeV2joule',
    'Quad',
    'Scatter',
    'bunchgen',
    'classes',
    'conversions',
    'eV2joule',
    'elegant_sim',
    'epsilon_0',
    'gamma2GeV',
    'gaussbunch_twiss',
    'scatter',
    'warnings',
    ]

from .driftmat import Drift
from .bend import Bend
from .quadmat import Quad
from .focusmat import Focus
from .beamline import Beamline
from .scatter import Scatter
from .beamparams import BeamParams
from .conversions import *      # noqa
from .elegant_sim_mod import *  # noqa
from .bunchgen import *         # noqa
from .classes import *          # noqa
from . import PWFA



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import numpy as _np

from .driftmat import driftmat
from .baseclass import baseclass


class Quad(baseclass):
    """
    Represents a quadrupole element.

    Parameters
    ----------
    
    length : float
        The length of the magnet.
    K1 : float
        Geometric focusing strength of the quad.
    order : float
        Order to calculate matrix elements.
    name : str
        Unique name to identify the element.
    """
    def __init__(self, length=0, K1=0, order=1, name=None, **kwargs):
        self.name   = name
        self._order  = int(order)
        self._type   = 'quad'
        self._length = _np.float64(length)
        self.K1      = _np.float64(K1)
        self._kwargs = kwargs

    @property
    def K1(self):
        """
        The geometric focusing strength.
        """
        return self._K1

    @K1.setter
    def K1(self, val):
        self._K1 = val

    @property
    def R(self):
        """
        The transfer matrix R for the quad.
        """
        return quadmat(L=self._length, K1=self.K1, order=self._order)

    @property
    def length(self):
        """
        The length of the quad.
        """
        return self._length

    def change_E(self, old_gamma, new_gamma):
        """
        Scale the setpoint energy of the magnet from the old energy old_gamma to new energy new_gamma.

        The best way to think about this is that the angle will change with different beam energy, so changing the beam energy changes the magnet’s properties, because the magnet’s B-field stays the same.
        """
        old_gamma = _np.float64(old_gamma)
        new_gamma = _np.float64(new_gamma)
        self.K1 *= old_gamma / new_gamma

    @property
    def ele_name(self):
        """
        Returns the elegant-styled name.
        """
        name = 'QUAD_{}_{:03.0f}'.format(self.name, self.ind)
        return name

    @property
    def ele_string(self):
        """
        Returns the full elegant entry.
        """
        string = '{}\t:KQUAD, L = {}, K1 = {}'.format(self.ele_name, self.length, self.K1)
        string = string + self._kwargs_str

        return string


def quadmat(K1=0, L=0, order=1):
    if ( K1 == 0 ):
        R = driftmat(L, order)
    else:
        rtK = _np.sqrt(_np.abs(K1))
        rtK_L = rtK * L
        sin_rtK_L = _np.sin(rtK_L)
        cos_rtK_L = _np.cos(rtK_L)
        sinh_rtK_L = _np.sinh(rtK_L)
        cosh_rtK_L = _np.cosh(rtK_L)
        R_f = _np.array(
            [[ cos_rtK_L    , sin_rtK_L/rtK ],
            [  -rtK*sin_rtK_L , cos_rtK_L   ]]
            )
        R_d = _np.array(
            [[ cosh_rtK_L   , sinh_rtK_L/rtK ],
            [  rtK*sinh_rtK_L , cosh_rtK_L   ]]
            )
        if ( K1 > 0 ):
            R = _np.zeros([6, 6])
            R[0:2, 0:2] = R_f
            R[2:4, 2:4] = R_d
            R[4:6, 4:6] = _np.identity(2)
        elif ( K1 < 0 ):
            R = _np.zeros([6, 6])
            R[0:2, 0:2] = R_d
            R[2:4, 2:4] = R_f
            R[4:6, 4:6] = _np.identity(2)
        else:
            print('Uhm who said what now??')
            print(K1)

    return R



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
from .baseclass import baseclass as _baseclass

__all__ = ['Drift']


class Drift(_baseclass):
    """
    Represents a drift element.

    Parameters
    ----------
    
    length : float
        The length of the drift.
    order : float
        Order to calculate matrix elements.
    name : str
        Unique name to identify the element.
    """
    def __init__(self, length=0, order=1, name=None, **kwargs):
        self.name   = name
        self._type   = 'drift'
        self._length = _np.float64(length)
        self._order  = int(order)
        self._kwargs = kwargs

    # Define transfer matrix property R
    @property
    def R(self):
        """
        The first-order transfer matrix.
        """
        return driftmat(self._length, self._order)

    def change_E(self, old_gamma, new_gamma):
        """
        Scale the energy of the drift. (Does nothing in a drift.)
        """
        pass

    @property
    def length(self, verbose=False):
        """
        The length of the drift.
        """
        if verbose:
            print('This is a drift matrix')
        return self._length

    @length.setter
    def length(self, value):
        print('Length changing for element ({}) from {} to {}'.format(self.name, self._length, value))
        self._length = _np.float64(value)

    @property
    def ele_name(self):
        """
        Returns the elegant-styled name.
        """
        name = 'CSRD_{}_{:03.0f}'.format(self.name, self.ind)
        return name

    @property
    def ele_string(self):
        """
        Returns the full elegant entry.
        """
        string = '{}\t:CSRDRIF,L={},USE_STUPAKOV=1,N_KICKS=1'.format(self.ele_name, self.length)
        string = string + self._kwargs_str

        return string


def driftmat(l=0, order=1):
    R_small = _np.array(
        [[ 1 , l ],
        [  0 , 1 ]]
        )
    R_small = _np.float_(R_small)

    R = _np.zeros([6, 6])
    R[0:2, 0:2] = R[2:4, 2:4] = R_small
    R[4:6, 4:6] = _np.float_(_np.identity(2))
    return R



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import numpy as _np
    import slactrac as _sltr
    import periodictable as _pt
    import scipy.constants as _spc
else:
    import unittest.mock as _mock
    _pt = _mock.MagicMock()


class Plasma(object):
    """
    A class for relating plasma density to plasma frequency :math:`\\omega_p` and ion focusing force.

    Input plasma density either `n_p` in SI units, or `n_p_cgs` in CGS units.

    Parameters
    ----------

    n_p : float
        Plasma density in SI units.
    n_p_cgs : float
        Plasma density in CGS units.
    species : periodictable.core.Element
        The species of gas used (see :class:`periodictable.core.Element`).
    """
    def __init__(self, n_p=None, n_p_cgs=None, species=_pt.hydrogen):
        if species is None:
            species = _pt.H
        # ============================
        # Reconcile n_p, n_p_cgs
        # ============================
        if n_p is None and n_p_cgs is None:
            raise ValueError('Keywords n_p and n_p_cgs cannot both be None')
        elif n_p is not None and n_p_cgs is not None:
            raise ValueError('Keywords n_p and n_p_cgs cannot both be specified')
        elif n_p is not None:
            self.n_p = n_p
        elif n_p_cgs is not None:
            self.n_p_cgs = n_p_cgs

        if type(species) != _pt.core.Element:
            raise TypeError('Input species must be derived from periodictable package.')
        else:
            self._species = species

    @property
    def species(self):
        """
        Species used in plasma.
        """
        return self._species

    @property
    def m(self):
        """
        Mass of the ion in AMU.
        """
        return self._species.mass * _spc.m_u

    @property
    def E_rest(self):
        """
        Rest energy of the plasma ion.
        """
        return self.m * _spc.c**2

    @property
    def n_p(self):
        """
        Plasma density in SI units.
        """
        return self._n_p

    @n_p.setter
    def n_p(self, n_p):
        self._n_p = n_p

    @property
    def w_p(self):
        """
        Plasma frequency :math:`\\omega_p` for given plasma density
        """
        return _np.sqrt(self.n_p * _np.power(_spc.e, 2) / (_spc.m_e * _spc.epsilon_0))

    def k_ion(self, E):
        """
        Geometric focusing force due to ion column for given plasma density as a function of *E*
        """
        return self.n_p * _np.power(_spc.e, 2) / (2*_sltr.GeV2joule(E) * _spc.epsilon_0)

    def beta_matched(self, E):
        return 1/_np.sqrt(self.k_ion(E))

    @property
    def n_p_cgs(self):
        """
        Plasma density in CGS units
        """
        return self.n_p * 1e-6

    @n_p_cgs.setter
    def n_p_cgs(self, n_p_cgs):
        self.n_p = n_p_cgs * 1e6



import os as _os
on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not on_rtd:
    import numpy as _np
    import scipy.constants as _spc


def gamma2GeV(gamma):
    """
    Finds the energy in GeV of an electron with relativistic Lorentz factor :math:`\\gamma`.

    Parameters
    ----------

    gamma : float
        Relativistic Lorentz factor :math:`\\gamma`.
    """
    n_pts = _np.size(gamma)
    if n_pts > 1:
        for i, val in enumerate(gamma):
            gamma[i] = gamma2GeV(val)
        return gamma
    else:
        gamma = _np.float(gamma)
        return gamma * ( _spc.m_e * (_spc.c**2) / (_spc.eV * _spc.giga) )


def GeV2gamma(E):
    """
    Finds the relativistic Lorentz factor :math:`\\gamma` of an electron from a given energy.

    Parameters
    ----------

    E : float
        Electron energy in GeV.
    """
    if _np.size(E) == 1:
        E = _np.float(E)
    else:
        E = _np.array(E, dtype='float')
    return E / ( _spc.m_e * (_spc.c**2) / (_spc.eV * _spc.giga) )


def GeV2joule(E):
    """
    Converts energy from units of GeV to joules.

    Parameters
    ----------

    E : float
        Energy in GeV.
    """
    return eV2joule(E) * _spc.giga


def eV2joule(E):
    """
    Converts energy from units of eV to joules.

    Parameters
    ----------

    E : float
        Energy in eV.
    """
    if _np.size(E) == 1:
        E = _np.float(E)
    else:
        E = _np.array(E, dtype='float')

    return E * _spc.eV



__all__ = [
    'RoundBeam',
    'EllipseBeam',
    'Ions1D',
    'Ions2D',
    'Match',
    'MatchPlasma',
    'Plasma',
    'GaussPartBeam',
    'SimBeam'
    ]
__all__.sort()

from .ions1d import Ions1D
from .ions2d import Ions2D
from .match import *      # noqa
from .plasma import *     # noqa
from .beam import *       # noqa
from .particles import *  # noqa
from .sim import SimBeam



import os as _os
_on_rtd = _os.environ.get('READTHEDOCS', None) == 'True'
if not _on_rtd:
    import numpy as _np

__all__ = ['Bunch']


class Bunch(object):
    """
    A class representing a bunch of particles. (Unfinished)
    """
    def __init__(self, x, xp, y, yp, z, delta, parts=None):
        self._parts = _np.zeros(6, _np.size(x))

    @property
    def x(self):
        """
        An array of x coordinates for particles.
        """
        return self._parts[0, :]

    @x.setter
    def x(self, val):
        self._parts[0, :] = val

    def prop(self, beamline, gamma=None):
        """
        Propagates the bunch down a *beamline*. (Unfinished)
        """
        if gamma is None:
            gamma = beamline.gamma



class baseclass(object):
    """
    Defines a basic beamline element. All beamline elements should have a *length*, an *order*, and a *name* to identify them.
    """
    def __init__(self, length=0, order=1, name=None):
        self.ind     = None
        self._name    = name
        self._length = length
        self._order  = order
        self._type   = None
        self._ele_type = None

    def _not_allowed(self, *args, **kwargs):
        raise AttributeError('This attribute is not available for element type: {}'.format(self._type))

    K1        = property(_not_allowed, _not_allowed)
    angle     = property(_not_allowed, _not_allowed)
    rotate    = property(_not_allowed, _not_allowed)
    thickness = property(_not_allowed, _not_allowed)
    radlength = property(_not_allowed, _not_allowed)

    @property
    def _kwargs_str(self):
        string = ''
        for kname, kvalue in self._kwargs.items():
            string = string + ', {} = {}'.format(kname, kvalue)

        return string

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, namestr):
        self._name = namestr

    @property
    def order(self):
        return self._order
